from qutip import * 
import numpy as np

from scipy.linalg.decomp import eigvals_banded
import scipy.sparse as spr


def evolve(H, rho0, tlist, c_ops, vectorize=False):
    r = mesolve(H, rho0, tlist, c_ops, options=Options(store_final_state=True, store_states=False))
    v = r.final_state
    return v.full().reshape((v.shape[0]**2,)) if vectorize else v


def generate_condition(negvls):
    if negvls==0:
        def condition(vv, vals):
            return np.abs(vv)>0.95
    else:
        NN=negvls
        def condition(vv, vals):
            index=np.argsort(abs(vals))
            return vv in vals[index[-NN:]]
    return condition



def AL(H, c_ops, rho0 ,n ,T, nsteps=100, negvls='steady_state', tau=10**(-3), check_first=100, check_every=20, eps=1e-12):
    
    #=================  PREPARATION  =================#
    def do_liouvillian(rho):
        HH = H.full()
        cc_ops = [cc.full() for cc in c_ops]
        Lrho = -1.j*(np.dot(HH,rho) - np.dot(rho,HH))
        for jump in cc_ops:
            Lrho = Lrho + np.dot(jump,np.dot(rho,np.conj(np.transpose(jump))))- 0.5*(np.dot(rho,np.dot(np.conj(np.transpose(jump)),jump))+ np.dot(np.dot(np.conj(np.transpose(jump)),jump),rho))
        return Lrho
        
    v = operator_to_vector(rho0).full()[:,0]
    v/=np.linalg.norm(v)
    m = v.shape[0]
    sqrtm = int(np.sqrt(m))

    E = np.zeros((n + 1, n), dtype=np.complex128)
    K = np.zeros((m, n+1), dtype=np.complex128)
    K[:, 0] = v  # Use it as the first Krylov vector

    
    #=================  ITERATION LOOP  =================#
    tlist=np.linspace(0, T, nsteps)
    condition = generate_condition(negvls)
    for k in range(n):
        v_obj = Qobj(v.reshape(rho0.shape), dims=rho0.dims)
        v = evolve(H, v_obj, tlist, c_ops, vectorize=True)

        #Gram-Schmidt orthogonalization
        for j,e in enumerate(K[:,0:k+1].T):
            E[j,k] = np.dot(e.conj(),v)
            v -= E[j,k]*e
        E[k+1,k] = np.linalg.norm(v)
        
        #Add the produced vector to the list, unless the zero vector is produced.
        if E[k+1,k] > eps:
            v /= E[k+1,k] #Ortonormalization
            K[:,k+1] = v  
        else: break
        
        
        #=================  CONVERGENCE CHECK  =================#
        convergence=0
        if k%check_every==0 and k>check_first:
            vals_eff, vecs_eff=np.linalg.eig(E[0:k, 0:k]) #calculate the effective eigenvalues and eigenvectors

            vals_good=[]
            vecs_good=[]
            vals_liouv_good=[]
            convergence=1
            for jj in range(len(vals_eff)): #loop to see which eigenvalues/ eigenvectors have converged
                if condition(vals_eff[jj], vals_eff)==True:
                    
                    vec=vecs_eff[:,jj]
                    vec=np.dot(K[:,0:k],vec) #transforms the vector in the effective basis to one expressed in the original basis
                    vec=vec/np.linalg.norm(vec)
                    
                    vec_obj=Qobj(vec.reshape((sqrtm,sqrtm)), dims=rho0.dims)
                    vec=evolve(H, vec_obj, tlist, c_ops, vectorize=False).full()
                    
                    diff=vals_eff[jj]*vec_obj.full() - vec
                    if np.linalg.norm(diff)<tau: #convergence measure
                        convergence*=1
                        vals_good.append(vals_eff[jj]) #appends converged eigenvalue of the evolution operator
                        vec/=np.linalg.norm(vec) 
                        Lv=do_liouvillian(vec) 
                        vals_liouv_good.append(np.dot(np.conj(vec).reshape((m,)), Lv.reshape((m,))))   #calculates and appends the eigenvalue of the Liouvillian
                        vecs_good.append(vec) #appends the converged eigenvector
                    else:
                        convergence*=0
        
        if convergence: break
    print('Total number of iterations = '+str(k))
    return K, E, vals_good, vecs_good, vals_liouv_good