from qutip import *
import numpy as np
import matplotlib.pyplot as plt
from IPython.display import HTML
from matplotlib import animation
import time
import torch
import pickle
from scipy.signal import savgol_filter as sgf

Hz  = 1e-9
GHz = 1e9*Hz
MHz = 1e6*Hz
kHz = 1e3*Hz

def promote(O):
    return sprepost(O,O.dag())

class parameters():
    def __init__(self, omegas=7.216, omegaq=5.938, chi=1.789, gamma1=1/8, gamma2=1/5, k1=1/55, nth=0.02):
        self.ωs = (2*np.pi) * omegas * GHz
        self.ωq = (2*np.pi) * omegaq * GHz
        self.Δ  = self.ωs-self.ωq
        self.χ  = (2*np.pi) * chi * MHz
        self.g  = np.sqrt(self.χ*self.Δ)
        self.λ = self.g/self.Δ
        self.χqs = 2*self.χ
        with np.errstate(divide='ignore'): 
            self.τ = np.divide(np.pi,self.χqs)
        self.Γ    = gamma1*MHz
        self.Γϕ   = gamma2*MHz
        self.Γtot = self.Γ/2 + self.Γϕ
        
        self.k1  = k1*MHz
        self.nth = nth
    
class CQED():
    def __init__(self, N, p):
        self.N  = N
        self.p  = p
        
        #cavity-only operatos
        self.c = destroy(N)
        self.cparity = (1j*np.pi*self.c.dag()*self.c).expm()
        
        #Full system operators
        self.a  = tensor(destroy(N), identity(2))
        self.σ  = tensor(identity(N), destroy(2))
        self.σz = tensor(identity(N), sigmaz())
        self.σy = tensor(identity(N), sigmay())
        self.σx = tensor(identity(N), sigmax())
        self.parity = (1j*np.pi*self.a.dag()*self.a).expm()
        
        #Full system rotations
        self.ryp = qutip.qip.operations.ry(np.pi/2)
        self.rym = qutip.qip.operations.ry(-np.pi/2)
        self.Ryp = tensor(identity(N),self.ryp)
        self.Rym = tensor(identity(N),self.rym)

        #Full system projectos
        self.Πg = tensor(identity(N), ket2dm(basis(2, 0)))
        self.Πe = tensor(identity(N), ket2dm(basis(2, 1)))
        self.Πp = tensor(identity(N), ket2dm(self.ryp*basis(2, 0)))
        self.Πm = tensor(identity(N), ket2dm(self.rym*basis(2, 0)))

        #superoperators
        self.Πp_super  = promote(self.Πp)
        self.Πe_super  = promote(self.Πe)
        self.Πg_super  = promote(self.Πg)
        self.σx_super  = promote(self.σx)
        self.Rym_super = promote(self.Rym)
        self.Ryp_super = promote(self.Ryp)
        
        #Noise terms
        k1_up = p.nth*p.k1
        k1_down = (p.nth+1)*p.k1
        self.c_ops = [np.sqrt(p.Γ)*self.σ , np.sqrt(p.Γtot/2)*self.σz, np.sqrt(k1_down)*self.a, np.sqrt(k1_up)*self.a.dag()]
        
        #Hamiltonians
        self.Hideal  = p.χ*self.σz*self.a.dag()*self.a 
        self.Happrox = p.Δ/2*self.σz + p.g*(self.σ.dag()*self.a + self.σ*self.a.dag())
        self.Happrox_fixq = (p.Δ+p.χ)*self.a.dag()*self.a + p.g*(self.σ.dag()*self.a + self.σ*self.a.dag()) - p.χ/2*self.σz
        
        #cavity only hamiltonians and noise
        self.closs = [np.sqrt(k1_down)*self.c, np.sqrt(k1_up)*self.c.dag()] #cavity loss
        self.Hcavity = p.ωq * self.c.dag()*self.c
        
    def get_rotated_projectors(self,res,tf=None):
        if type(res)==qutip.qobj.Qobj:
            if tf==None:
                print("Error:Specify a time")
                return 0,0
            else:
                ξ = -(self.p.Δ+self.p.χ)*tf
                sx,sy = expect(self.σx,res), expect(self.σy,res)
        elif hasattr(res,'final_state'):
            sx,sy = expect(self.σx,res.final_state), expect(self.σy,res.final_state)
            ξ = -(self.p.Δ+self.p.χ)*res.times[-1]
        else:
            ξ = -(self.p.Δ+self.p.χ)*res.times 
            sx,sy = (res.expect[0], res.expect[1]) if res.num_expect else (expect(self.σx,res.states), expect(self.σy,res.states))
        
        Pip = 0.5*(1+np.cos(ξ)*sx - np.sin(ξ)*sy)
        Pim = 0.5*(1-np.cos(ξ)*sx + np.sin(ξ)*sy)
        return Pip,Pim
        
        
    def get_cat(self,n,parity='even',phase=0.):
        alpha = np.sqrt(n)*np.exp(1.j*phase)
        cat = (coherent(self.N,alpha) + coherent(self.N,-alpha)) if parity=='even' else (coherent(self.N,alpha) - coherent(self.N,-alpha))
        return cat.unit()
    
    def xmeasure(self,rho):
        pp = np.real(expect(self.Πp,rho))
        c = np.random.rand()
        if c<pp:
            rhof = (self.Πp*rho).unit()
            out  = 1
            cav  = rhof.ptrace(0).unit()
        else:
            rhof = (self.Πm*rho).unit()
            out  = -1
            cav  = rhof.ptrace(0).unit()
        return out, rhof, cav
    
    def get_SingleShotFidelity(self, n, H, c_ops=None, ntimes=10000, tf=None, rotate=True):
        tlist = np.linspace(0,self.p.τ,int(ntimes)) if tf==None else np.linspace(0,tf,int(ntimes))
        ψq = self.ryp * fock(2,0)
        ψc_list = [self.get_cat(n,'odd'), self.get_cat(n,'even')]
        
        if not(rotate): projectors=[self.Πp, self.Πm]
        F = 2
        for i,ψc in enumerate(ψc_list):
            ψ   = tensor(ψc, ψq)
            res = mesolve(H,ψ,tlist,c_ops=c_ops,options=Options(store_states=False, store_final_state=True))
            remove = self.get_rotated_projectors(res)[i] if rotate else expect(projectors[i], res.final_state)
            F  -= remove
        return F/2
    
    
    
    