from qutip import *
import numpy as np

class dynamic_mesolve():
    def __init__(self, generateH, Hargs, rho0, nsteps, Δt, tmax, A, pillow=4, tresh=1e-5, fixdt=False):
        self.generateH= generateH
        self.pillow   = pillow
        self._nsteps  = nsteps
        self.fixdt    = fixdt
        self.tresh    = tresh
        self._args    = Hargs
        self.tmax     = tmax
        self._Δt      = Δt
        
        self._rho     = rho0 if rho0.type=='oper' else ket2dm(rho0)        
        self._N       = np.array(self._rho.dims[0], dtype='int32')
        self.sysnum   = self._N.size
        
        self._H, self._c_ops, self._e_ops, self.args = self.generateH(*self._N, *self._args) 
        self._solve(A)
  

    def _update_L(self):
        self._H, self._c_ops, self._e_ops, _ = self.generateH(*self._N, *self._args) 
        return None        
  
    
    def _check_rhos(self):
        rhos = self._rho.ptrace(0)
        Ncut = 1+int(np.argwhere( np.invert( np.all(np.abs(rhos.full())<self.tresh, axis=0) ) )[-1])
        
        if (self._N[0]-Ncut)>self.pillow:
            self._N[0] = Ncut + self.pillow
            self._rho = Qobj(self._rho[:self._N.prod(), :self._N.prod()], dims=[list(self._N)]*2)
        
        elif (self._N[0]-Ncut)<self.pillow:
            add = self.pillow - (self._N[0]-Ncut)
            self._N[0] += add
            padder = add*self._N[1:].prod()
            self._rho = Qobj( np.pad(self._rho.full(),((0, padder), (0, padder))), type='oper', dims=[list(self._N)]*2)
    
    
    def _check_rhor(self):
        rhor = self._rho.ptrace(1)
        Ncut = 1+int(np.argwhere( np.invert( np.all(np.abs(rhor.full())<self.tresh, axis=0) ) )[-1])

        if (self._N[1]-Ncut)>self.pillow:
            Nrnew = Ncut + self.pillow
            ilist = np.array([i for i in range(self._N.prod()) if i%self._N[1] in np.arange(Nrnew)])
            
            self._N[1] = Nrnew
            self._rho  = Qobj(self._rho.data[ilist,:][:,ilist], dims=[list(self._N)]*2)
            
        elif (self._N[1]-Ncut)<self.pillow:
            add = self.pillow - (self._N[1]-Ncut)
            ilist = np.array([i for i in range(self._N.prod()) if i%self._N[1]==0])
            self._rho = Qobj(np.block([[np.pad(self._rho[j:j+self._N[1],i:i+self._N[1]],(0,add)) \
                                        for i in ilist] for j in ilist] ))
            self._N[1] += add
            self._rho.dims = [list(self._N)]*2
            
        
    def _update_dynamics(self, r, A, t, dt):
        n1, n0 = np.array([r.expect[i][-1] for i in range(self.sysnum)]), np.array([r.expect[i][-2] for i in range(self.sysnum)])
        ndot   = np.array([np.abs(n0[i]-n1[i])/dt for i in range(self.sysnum)])
        
        self._Δt = np.real(np.min([A*np.max([(n1/ndot).max(),1/ndot.max()]), self.tmax-t])) if not(self.fixdt) else self._Δt

        self._rho = r.final_state
        self._check_rhos()
        if self.sysnum>1: self._check_rhor()
        
        self._nsteps = int(np.max([np.round(ndot.max()*self._Δt)*10, 20]))
        return None
        
        
    def _solve(self,A):
        self.expect      = [np.array([])]*len(self._e_ops)
        self.times       = np.array([])
        self.jump_times  = []
        self.jump_Ncut   = []
        
        t = 0
        while(t<self.tmax):
            self.jump_times.append(t)
            self.jump_Ncut.append(np.copy(self._N))
            
            Δtlist, dt = np.linspace(0, self._Δt, self._nsteps, retstep=True)
            r = mesolve(self._H, self._rho, t+Δtlist, self._c_ops, self._e_ops,\
                        options=Options(store_final_state=True), args=self.args)
            
            self.expect = [np.concatenate((self.expect[i],r.expect[i])) for i in range(len(self._e_ops))]
            self.times  = np.concatenate((self.times,(t+Δtlist)))
            t += self._Δt
            self._update_dynamics(r, A, t, dt)
            self._update_L()
        self.jump_Ncut = np.array(self.jump_Ncut)
        for var in [self._H, self._c_ops, self._e_ops, self._rho, self._N, self._Δt, self._nsteps,\
                    self.generateH, self._args, r]:
            del var
            
            
            
            
