import numpy as np
from qutip import *
from tabulate import tabulate
from scipy.sparse.linalg import splu




def print_params(params):
    Us  = params['Us']
    Ur  = params['Ur']
    G   = params['G']
    η   = params['η']
    g2  = params['g2']
    χrs = params['χrs']
    ϵd  = params['εd']
    κr  = params['κr']
    ξp  = params['ξp']
    
    floatfmt = ".2e"
    
    title   = dict(Dimensional_Vars='Value [MHz]')
    #toprint = dict(Us=Us/(2*np.pi), Ur=Ur/(2*np.pi), G=G/(2*np.pi), χrs=χrs/(2*np.pi),\
    #               η=η/(2*np.pi), κr=κr/(2*np.pi))
    toprint = dict(Us=Us, Ur=Ur, G=G, χrs=χrs, η=η, κr=κr)
    title.update(toprint)
    print(tabulate(title.items(), headers='firstrow', floatfmt=floatfmt, tablefmt="fancy_grid", numalign="decimals"))
    print("\n")
    
    
    title   = dict(Dimensionless_Vars='Value [η]')
    toprint = dict(Us=Us/η, G=G/η, χrs=χrs/η, κr = κr/η)
    title.update(toprint)
    print(tabulate(title.items(), headers='firstrow', floatfmt=floatfmt, tablefmt="fancy_grid", numalign="decimals"))
    print("\n")
    

    title   = dict(Adiabaticity_conditions='Value [κr]')
    toprint = dict(Us=Us/κr, g2=g2/κr, ϵd=ϵd/κr, χrs=χrs/κr)
    title.update(toprint)
    print(tabulate(title.items(), headers='firstrow', floatfmt=floatfmt, tablefmt="fancy_grid", numalign="decimals"))
    print("\n")
    
    A = κr/(4*Ur*ξp**2)
    B = κr/Us
    ν = ϵd/κr
    n = 2*ν*np.sqrt(B/(A+1/A))
    
    title   = dict(Independent_pars='Value')
    toprint = dict(A=A, n=n, ν=ν, ξp=ξp, ξp2=ξp**2)
    title.update(toprint)
    print(tabulate(title.items(), headers='firstrow', floatfmt=".4f", tablefmt="fancy_grid", numalign="decimals"))
    
    return 




def get_ss_fromblock(k, parity):
    A = k.Lsymm[parity].data.tocsc()
    m = A.shape[0]
    msqrt = int(np.sqrt(m))
    
    b = np.ones(m)*1e-12
    B = splu(A)
    x = B.solve(b)
    
    ss = Qobj(x.reshape((msqrt,msqrt), order='F'))
    ss = k.full_completion(operator_to_vector(ss),parity)
    return ss


def get_fidelity(ss, get_entropy=False):
    N = ss.shape[0]
    a = destroy(N)
    parity = (1j*np.pi*a.dag()*a).expm()
    
    p = np.round(expect(parity, ss),2)
    α = np.sqrt(expect(a**2,ss))
    
    if p==1.:
        ss_est = ket2dm( (coherent(N, α) + coherent(N, -α)).unit() ) 
    elif p==-1.:
        ss_est = ket2dm( (coherent(N, α) - coherent(N, -α)).unit() )
    else:
        ss_est = 0
        print("Error: ss is not an eigenstate of the parity operator.")
        return 
    f = ( (ss.sqrtm() * ss_est * ss.sqrtm()).sqrtm() ).tr()
    if get_entropy : 
        s = entropy_vn(ss)
        return f.real, s
    else:
        return f.real