import numpy as np
from qutip import * 
from qutip.solver import Result
from qutip.ui.progressbar import TextProgressBar, BaseProgressBar


def check_psis(psi,thresh,pillow):
    N = np.array(psi.dims[0])
    sysnum = len(N)
    if sysnum==1:
        Ncut = 1+int(np.argwhere( np.invert( np.abs(psi.full().ravel("F"))<thresh ) )[-1])
    else:
        rho  = psi.ptrace(0)
        Ncut = 1+int(np.argwhere( np.invert( np.all(np.abs(rho.full())<thresh, axis=0) ) )[-1])

    if (N[0]-Ncut)>pillow:
        N[0] = Ncut + pillow
        psi = Qobj(psi[:N.prod()], type='ket', dims=[list(N),[1]*sysnum])

    elif (N[0]-Ncut)<pillow:
        add = pillow - (N[0]-Ncut)
        N[0] += add
        padder = add*N[1:].prod()
        psi = Qobj( np.pad(psi.full().ravel("F"),(0, padder)), type='ket', dims=[list(N),[1]*sysnum])
    return psi, N


def check_psir(psi,thresh,pillow):
    N = np.array(psi.dims[0])
    sysnum = len(N)
    rho  = psi.ptrace(1)
    Ncut = 1+int(np.argwhere( np.invert( np.all(np.abs(rho.full())<thresh, axis=0) ) )[-1])

    if (N[1]-Ncut)>pillow:
        Nrnew = Ncut + pillow
        ilist = np.array([i for i in range(N.prod()) if i%N[1] in np.arange(Nrnew)])

        N[1] = Nrnew
        psi  = Qobj(psi[ilist], dims=[list(N),[1]*sysnum])

    elif (N[1]-Ncut)<pillow:
        add = pillow - (N[1]-Ncut)
        ilist = np.array([i for i in range(N.prod()) if i%N[1]==0])
        N[1] += add
        psi = Qobj(np.hstack([np.pad(psi[i:i+N[1]-add].ravel("F"), (0,add)) for i in ilist]), dims = [list(N),[1]*sysnum])
    return psi, N

def update_output(out,r):
    out.expect    = [np.concatenate((out.expect[i],r.expect[i])) for i in range(out.num_expect)]
    out.times     = np.concatenate((out.times,r.times))
    out.col_times = np.concatenate((out.col_times, r.col_times[0]))
    out.col_which = np.concatenate((out.col_which, r.col_which[0]))

def reset_output(out, e_ops, c_ops):
    out.num_expect   = len(e_ops)
    out.num_collapse = len(c_ops)
    out.times        = np.array([])
    out.expect       = [np.array([])]*out.num_expect
    out.col_times    = np.array([])
    out.col_which    = np.array([])

def merge_out(outlist, average_expect=True):
    out = Result()
    out.times = outlist[0].times
    out.ntraj = outlist[0].ntraj
    out.num_expect = outlist[0].num_expect
    out.num_collapse = outlist[0].num_collapse
    if average_expect:
        out.expect = [np.array([r.expect[i] for r in outlist]).mean(0) for i in range(out.num_expect)]
    else:
        out.expect = [np.array([r.expect[i] for r in outlist]) for i in range(out.num_expect)]
    
    out.col_times = [r.col_times for r in outlist]
    out.col_which = [r.col_which for r in outlist]
    return out
    
def single_dyntraj(psi, generateL, tmax, steps, gamma, jumps, pillow, thresh):
    N = np.array(psi.dims[0])
    H, c_ops, e_ops, args = generateL(*N)
    
    t  = 0
    Δn = steps//jumps 
    tlist = np.linspace(0,tmax**(1/gamma),steps)**gamma
    
    out = Result() 
    reset_output(out, e_ops, c_ops)
     
    for i in range(1,jumps+1):
        tev = tlist[i*Δn:(i+1)*Δn+1]
        r   = mcsolve(H, psi, tev, c_ops, e_ops, ntraj=1, args=args, progress_bar=False,\
                    options=Options(num_cpus=1, store_final_state=True))
        update_output(out,r)
        
        psi = Qobj(r.final_state[0].full(), dims=[list(N),[1]*len(N)])
        psi, N = check_psis(psi, thresh, pillow)
        psi, N = check_psir(psi, thresh, pillow)

        H, c_ops, e_ops, args = generateL(*N)
    return out

def dynamic_mcsolve(generateL, tmax, steps, gamma, jumps, psi0, ntraj, pillow=2, thresh=1e-4, progress_bar=False):

    progress_bar = BaseProgressBar() if not(progress_bar) else TextProgressBar()
    
    map_kwargs = {}
    map_kwargs_= {'progress_bar': progress_bar}
    map_kwargs_.update(map_kwargs)
    map_kwargs = map_kwargs_

    psi0list = [psi0]*ntraj
    res = parallel_map(single_dyntraj, psi0list, task_args=(generateL, tmax, steps, gamma, jumps, pillow, thresh), **map_kwargs)
    out = merge_out(res)
    out.ntraj = ntraj
    return out